import asyncio
import io
import logging
import ssl

import aiohttp
from aiohttp import web

from amqtt.adapters import ReaderAdapter, WriterAdapter
from amqtt.broker import Broker
from amqtt.contexts import BrokerConfig, ListenerConfig, ListenerType

logger = logging.getLogger(__name__)

MQTT_LISTENER_NAME = "myMqttListener"

async def hello(request):
    """Get request handler"""
    return web.Response(text="Hello, world")

class WebSocketResponseReader(ReaderAdapter):
    """Interface to allow mqtt broker to read from an aiohttp websocket connection."""

    def __init__(self, ws: web.WebSocketResponse):
        self.ws = ws
        self.buffer = bytearray()

    async def read(self, n: int = -1) -> bytes:
        """Read 'n' bytes from the datastream, if < 0 read all available bytes

        Raises:
            BrokerPipeError : if reading on a closed websocket connection

        """
        # continue until buffer contains at least the amount of data being requested
        while not self.buffer or len(self.buffer) < n:
            # if the websocket is closed
            if self.ws.closed:
                raise BrokenPipeError

            try:
                # read from stream
                msg = await asyncio.wait_for(self.ws.receive(), timeout=0.5)
                # mqtt streams should always be binary...
                if msg.type == aiohttp.WSMsgType.BINARY:
                    self.buffer.extend(msg.data)
                elif msg.type == aiohttp.WSMsgType.CLOSE:
                    raise BrokenPipeError

            except asyncio.TimeoutError:
                raise BrokenPipeError

        # return all bytes currently in the buffer
        if n == -1:
            result = bytes(self.buffer)
            self.buffer.clear()
        # return the requested number of bytes from the buffer
        else:
            result = self.buffer[:n]
            del self.buffer[:n]

        return result

    def feed_eof(self) -> None:
        pass

class WebSocketResponseWriter(WriterAdapter):
    """Interface to allow mqtt broker to write to an aiohttp websocket connection."""

    def __init__(self, ws: web.WebSocketResponse, request: web.Request):
        super().__init__()
        self.ws = ws

        # needed for `get_peer_info`
        # https://docs.python.org/3/library/socket.html#socket.socket.getpeername
        peer_name = request.transport.get_extra_info("peername")
        if peer_name is not None:
            self.client_ip, self.port = peer_name[0:2]
        else:
            self.client_ip, self.port = request.remote, 0

        # interpret AF_INET6
        self.client_ip = "localhost" if self.client_ip == "::1" else self.client_ip

        self._stream = io.BytesIO(b"")

    def write(self, data: bytes) -> None:
        """Add bytes to stream buffer."""
        self._stream.write(data)

    async def drain(self) -> None:
        """Send the collected bytes in the buffer to the websocket connection."""
        data = self._stream.getvalue()
        if data and len(data):
            await self.ws.send_bytes(data)
        self._stream = io.BytesIO(b"")

    def get_peer_info(self) -> tuple[str, int] | None:
        return self.client_ip, self.port

    async def close(self) -> None:
        # no clean up needed, stream will be gc along with instance
        pass

    def get_ssl_info(self) -> ssl.SSLObject | None:
        pass


async def mqtt_websocket_handler(request: web.Request) -> web.StreamResponse:

    # establish connection by responding to the websocket request with the 'mqtt' protocol
    ws = web.WebSocketResponse(protocols=["mqtt"])
    await ws.prepare(request)

    # access the broker created when the server started
    b: Broker = request.app["broker"]

    # hand-off the websocket data stream to the broker for handling
    # `listener_name` is the same name of the externalized listener in the broker config
    await b.external_connected(WebSocketResponseReader(ws), WebSocketResponseWriter(ws, request), MQTT_LISTENER_NAME)

    logger.debug("websocket connection closed")
    return ws


async def websocket_handler(request: web.Request) -> web.StreamResponse:
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    async for msg in ws:
        logging.info(msg)

    logging.info("websocket connection closed")
    return ws

def main():
    # create an `aiohttp` server
    lp = asyncio.get_event_loop()
    app = web.Application()
    app.add_routes(
        [
            web.get("/", hello), # http get request/response route
            web.get("/ws", websocket_handler), # standard websocket handler
            web.get("/mqtt", mqtt_websocket_handler), # websocket handler for mqtt connections
        ])
    # create background task for running the `amqtt` broker
    app.cleanup_ctx.append(run_broker)

    # make sure that both `aiohttp` server and `amqtt` broker run in the same loop
    #  so the server can hand off the connection to the broker (prevents attached-to-a-different-loop `RuntimeError`)
    web.run_app(app, loop=lp)


async def run_broker(_app):
    """App init function to start (and then shutdown) the `amqtt` broker.
    https://docs.aiohttp.org/en/stable/web_advanced.html#background-tasks
    """
    # standard TCP connection as well as an externalized-listener
    cfg = BrokerConfig(
        listeners={
            "default":ListenerConfig(type=ListenerType.TCP, bind="127.0.0.1:1883"),
            MQTT_LISTENER_NAME: ListenerConfig(type=ListenerType.EXTERNAL),
        }
    )

    # make sure the `Broker` runs in the same loop as the aiohttp server
    loop = asyncio.get_event_loop()
    broker = Broker(config=cfg, loop=loop)

    # store broker instance so that incoming requests can hand off processing of a datastream
    _app["broker"] = broker
    # start the broker
    await broker.start()

    # pass control back to web app
    yield

    # closing activities
    await broker.shutdown()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



import asyncio
import unittest

from amqtt.adapters import BufferReader
from amqtt.mqtt.packet import PacketIdVariableHeader
from amqtt.mqtt.unsuback import UnsubackPacket


class UnsubackPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b"\xb0\x02\x00\x0a"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(UnsubackPacket.from_stream(stream))
        assert message.variable_header is not None
        assert message.variable_header.packet_id == 10

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        publish = UnsubackPacket(variable_header=variable_header)
        out = publish.to_bytes()
        assert out == b"\xb0\x02\x00\n"



from collections.abc import MutableMapping
from dataclasses import dataclass, fields, is_dataclass
import json
from typing import Any

from amqtt.contrib.shadows.states import MetaTimestamp, ShadowOperation, State, StateDocument


def asdict_no_none(obj: Any) -> Any:
    """Create dictionary from dataclass, but eliminate any key set to `None`."""
    if is_dataclass(obj):
        result = {}
        for f in fields(obj):
            value = getattr(obj, f.name)
            if value is not None:
                result[f.name] = asdict_no_none(value)
        return result
    if isinstance(obj, list):
        return [asdict_no_none(item) for item in obj if item is not None]
    if isinstance(obj, dict):
        return {
            key: asdict_no_none(value)
            for key, value in obj.items()
            if value is not None
        }
    return obj


def create_shadow_topic(device_id: str, shadow_name: str, message_op: "ShadowOperation") -> str:
    """Create a shadow topic for message type."""
    return f"$shadow/{device_id}/{shadow_name}/{message_op}"


class ShadowMessage:
    def to_message(self) -> bytes:
        return json.dumps(asdict_no_none(self)).encode("utf-8")


@dataclass
class GetAcceptedMessage(ShadowMessage):
    state: State[dict[str, Any]]
    metadata: State[MetaTimestamp]
    timestamp: int
    version: int

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.GET_ACCEPT)


@dataclass
class GetRejectedMessage(ShadowMessage):
    code: int
    message: str
    timestamp: int | None = None

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.GET_REJECT)


@dataclass
class UpdateAcceptedMessage(ShadowMessage):
    state: State[dict[str, Any]]
    metadata: State[MetaTimestamp]
    timestamp: int
    version: int

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.UPDATE_ACCEPT)


@dataclass
class UpdateRejectedMessage(ShadowMessage):
    code: int
    message: str
    timestamp: int

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.UPDATE_REJECT)


@dataclass
class UpdateDeltaMessage(ShadowMessage):
    state: MutableMapping[str, Any]
    metadata: MutableMapping[str, Any]
    timestamp: int
    version: int

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.UPDATE_DELTA)


class UpdateIotaMessage(UpdateDeltaMessage):
    """Same format, corollary name."""

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.UPDATE_IOTA)


@dataclass
class UpdateDocumentMessage(ShadowMessage):
    previous: StateDocument
    current: StateDocument
    timestamp: int

    @staticmethod
    def topic(device_id: str, shadow_name: str) -> str:
        return create_shadow_topic(device_id, shadow_name, ShadowOperation.UPDATE_DOCUMENTS)



"""Module for the shadow state plugin."""

from .plugin import ShadowPlugin, ShadowTopicAuthPlugin
from .states import ShadowOperation

__all__ = ["ShadowOperation", "ShadowPlugin", "ShadowTopicAuthPlugin"]



from dataclasses import dataclass, field
import logging

from passlib.context import CryptContext
from sqlalchemy.ext.asyncio import create_async_engine

from amqtt.broker import BrokerContext
from amqtt.contexts import Action
from amqtt.contrib.auth_db.managers import TopicManager, UserManager
from amqtt.contrib.auth_db.models import Base, PasswordHasher
from amqtt.errors import MQTTError
from amqtt.plugins.base import BaseAuthPlugin, BaseTopicPlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)


def default_hash_scheme() -> list[str]:
    """Create config dataclass defaults."""
    return ["argon2", "bcrypt", "pbkdf2_sha256", "scrypt"]


class UserAuthDBPlugin(BaseAuthPlugin):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)

        # access the singleton and set the proper crypt context
        pwd_hasher = PasswordHasher()
        pwd_hasher.crypt_context = CryptContext(schemes=self.config.hash_schemes, deprecated="auto")

        self._user_manager = UserManager(self.config.connection)
        self._engine = create_async_engine(f"{self.config.connection}")

    async def on_broker_pre_start(self) -> None:
        """Sync the schema (if configured)."""
        if not self.config.sync_schema:
            return
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def authenticate(self, *, session: Session) -> bool | None:
        """Authenticate a client's session."""
        if not session.username or not session.password:
            return False

        user_auth = await self._user_manager.get_user_auth(session.username)
        if not user_auth:
            return False

        return bool(session.password) and user_auth.verify_password(session.password)

    @dataclass
    class Config:
        """Configuration for DB authentication."""

        connection: str
        """SQLAlchemy connection string for the asyncio version of the database connector:

        - `mysql+aiomysql://user:password@host:port/dbname`
        - `postgresql+asyncpg://user:password@host:port/dbname`
        - `sqlite+aiosqlite:///dbfilename.db`
        """
        sync_schema: bool = False
        """Use SQLAlchemy to create / update the database schema."""
        hash_schemes: list[str] = field(default_factory=default_hash_scheme)
        """list of hash schemes to use for passwords"""


class TopicAuthDBPlugin(BaseTopicPlugin):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)

        self._topic_manager = TopicManager(self.config.connection)
        self._engine = create_async_engine(f"{self.config.connection}")

    async def on_broker_pre_start(self) -> None:
        """Sync the schema (if configured)."""
        if not self.config.sync_schema:
            return
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool | None:
        if not session or not session.username or not topic:
            return None

        try:
            topic_auth = await self._topic_manager.get_topic_auth(session.username)
            topic_list = getattr(topic_auth, f"{action}_acl")
        except MQTTError:
            return False

        return topic in topic_list

    @dataclass
    class Config:
        """Configuration for DB topic filtering."""

        connection: str
        """SQLAlchemy connection string for the asyncio version of the database connector:

        - `mysql+aiomysql://user:password@host:port/dbname`
        - `postgresql+asyncpg://user:password@host:port/dbname`
        - `sqlite+aiosqlite:///dbfilename.db`
        """
        sync_schema: bool = False
        """Use SQLAlchemy to create / update the database schema."""



import logging
from enum import Enum, auto

import pytest
import pytest_asyncio
from aiohttp import web
from aiohttp.web import Response

from amqtt.broker import BrokerContext, Broker
from amqtt.contexts import Action
from amqtt.contrib.http import UserAuthHttpPlugin, TopicAuthHttpPlugin, ParamsMode, ResponseMode, RequestMethod
from amqtt.session import Session

logger = logging.getLogger(__name__)


def determine_response(d, matcher_field) -> Response:
    """The request's parameters determine what response is expected."""
    if d['username'] == 'json':
        # special case, i_am_null respond with None
        if d[matcher_field] == 'i_am_null':
            return web.json_response({'Ok': None})
        # otherwise, respond depending on if username and client_id match
        return web.json_response({'Ok': d['username'] == d[matcher_field]})
    elif d['username'] == 'status':
        if d[matcher_field] == 'i_am_null':
            return web.Response(status=500)
        return web.Response(status=200) if d['username'] == d[matcher_field] else web.Response(status=400)
    else: # text
        return web.Response(text='ok' if d['username'] == d[matcher_field] else 'error')


@pytest.fixture
async def empty_broker():
    config = {'listeners': {'default': { 'type': 'tcp', 'bind': '127.0.0.1:1883'}}, 'plugins': {}}
    broker = Broker(config)
    yield broker


async def all_request_handler(request: web.Request) -> Response:
    """The url and method type is used determine how the data was passed to the request."""
    if 'form' in str(request.url):
        if request.method == 'GET':
            d = request.query
        else:
            d = dict(await request.post())
    else:
        d = dict(await request.json())

    if '/user' in str(request.url):
        matcher = 'password'
        assert 'username' in d
        assert 'password' in d
        assert 'client_id' in d
    else:
        assert 'username' in d
        assert 'client_id' in d
        assert 'topic' in d
        assert 'acc' in d
        assert 1 <= int(d['acc']) <= 4
        matcher = 'client_id'

    return determine_response(d, matcher)


@pytest_asyncio.fixture
async def http_server():
    app = web.Application()

    # create all the routes for the various configuration options to handle the various use cases
    routes = []
    for test_kind in ('user', 'acl'):
        for test_data in ('json', 'form'):
            for method in ('get', 'post', 'put'):
                func_method = getattr(web, method)
                routes.append(func_method(f'/{test_kind}/{test_data}', all_request_handler))

    app.add_routes(routes)

    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, "127.0.0.1", 8080)
    await site.start()

    yield f"http://127.0.0.1:8080"

    await runner.cleanup()


def test_server_up_and_down(http_server):
    pass


class TypeOfHttpTest(Enum):
    AUTH = auto()
    ACL = auto()


def generate_use_cases(kind: TypeOfHttpTest):
    # generate all variations of the plugin's configuration options for both auth and topic
    # e.g. (TestKind.AUTH, '/user/json', RequestMethod.GET, ParamsMode.JSON, ResponseMode.JSON, 'json', 'json', True),

    cases: list[tuple[str, str, RequestMethod, ParamsMode, ResponseMode, str, str, bool]] = []

    root_url = 'user' if kind == TypeOfHttpTest.AUTH else 'acl'
    for request in RequestMethod:
        for params in ParamsMode:
            for response in ResponseMode:
                url = f'/{root_url}/json' if params == ParamsMode.JSON else f'/{root_url}/form'
                for is_authenticated in [True, False, None]:
                    if is_authenticated is None:
                        pwd = 'i_am_null'
                    elif is_authenticated:
                        pwd = f'{response.value}'
                    else:
                        pwd = f'not{response.value}'

                    if response == ResponseMode.TEXT and is_authenticated is None:
                        is_authenticated = False

                    case = (kind, url, request, params, response, response.value, f"{pwd}", is_authenticated)
                    cases.append(case)
    return cases

def test_generated_use_cases():
    cases = generate_use_cases(kind=TypeOfHttpTest.AUTH)
    assert len(cases) == 54


@pytest.mark.parametrize("kind,url,request_method,params_mode,response_mode,username,matcher,is_allowed",
                         generate_use_cases(TypeOfHttpTest.AUTH))
@pytest.mark.asyncio
async def test_request_auth_response(empty_broker, http_server, kind, url,
                                     request_method, params_mode, response_mode,
                                     username, matcher, is_allowed):

    context = BrokerContext(broker=empty_broker)
    context.config = UserAuthHttpPlugin.Config(
        host="127.0.0.1",
        port=8080,
        user_uri=url,
        request_method=request_method,
        params_mode=params_mode,
        response_mode=response_mode,
    )
    http_acl = UserAuthHttpPlugin(context)
    logger.warning(f'kind is {kind}')

    session = Session()
    session.client_id = "my_client_id"
    session.username = username
    session.password = matcher
    assert await http_acl.authenticate(session=session) == is_allowed

    await http_acl.on_broker_pre_shutdown()


@pytest.mark.parametrize("kind,url,request_method,params_mode,response_mode,username,matcher,is_allowed",
                         generate_use_cases(TypeOfHttpTest.ACL))
@pytest.mark.asyncio
async def test_request_topic_response(empty_broker, http_server, kind, url,
                                     request_method, params_mode, response_mode,
                                     username, matcher, is_allowed):
    context = BrokerContext(broker=empty_broker)
    context.config = TopicAuthHttpPlugin.Config(
        host="127.0.0.1",
        port=8080,
        topic_uri=url,
        request_method=request_method,
        params_mode=params_mode,
        response_mode=response_mode,
    )
    http_acl = TopicAuthHttpPlugin(context)

    s = Session()
    s.username = username
    s.client_id = matcher
    t = 'my/topic'
    a = Action.PUBLISH
    assert await http_acl.topic_filtering(session=s, topic=t, action=a) == is_allowed

    await http_acl.on_broker_pre_shutdown()



    # if kind == TestKind.ACL:

    # else:


import asyncio
import unittest

from amqtt.adapters import BufferReader
from amqtt.mqtt.packet import PacketIdVariableHeader
from amqtt.mqtt.unsubscribe import UnsubscribePacket, UnubscribePayload


class UnsubscribePacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b"\xa2\x0c\x00\n\x00\x03a/b\x00\x03c/d"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(UnsubscribePacket.from_stream(stream))
        assert message.payload.topics[0] == "a/b"
        assert message.payload.topics[1] == "c/d"

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = UnubscribePayload(["a/b", "c/d"])
        publish = UnsubscribePacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        assert out == b"\xa2\x0c\x00\n\x00\x03a/b\x00\x03c/d"



import asyncio
import logging
from pathlib import Path
import unittest

import pytest

from amqtt.plugins.authentication import AnonymousAuthPlugin, FileAuthPlugin
from amqtt.contexts import BaseContext
from amqtt.plugins.base import BaseAuthPlugin
from amqtt.session import Session

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=formatter)


@pytest.mark.asyncio
async def test_base_no_config(logdog):
    """Check BaseTopicPlugin returns false if no topic-check is present."""
    with logdog() as pile:
        context = BaseContext()
        context.logger = logging.getLogger("testlog")
        context.config = {}

        plugin = BaseAuthPlugin(context)
        s = Session()
        authorised = await plugin.authenticate(session=s)
        assert authorised is False

        # Warning messages are only generated if using deprecated plugin configuration on initial load
        log_records = list(pile.drain(name="testlog"))
        assert len(log_records) == 1
        assert log_records[0].levelno == logging.WARNING
        assert log_records[0].message == "'auth' section not found in context configuration"


class TestAnonymousAuthPlugin(unittest.TestCase):
    def setUp(self) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()

    def test_allow_anonymous_dict_config(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {"auth": {"allow-anonymous": True}}
        s = Session()
        s.username = ""
        auth_plugin = AnonymousAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert ret

    def test_allow_anonymous_dataclass_config(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = AnonymousAuthPlugin.Config(allow_anonymous=True)
        s = Session()
        s.username = ""
        auth_plugin = AnonymousAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert ret

    def test_disallow_anonymous(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {"auth": {"allow-anonymous": False}}
        s = Session()
        s.username = ""
        auth_plugin = AnonymousAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert not ret

    def test_allow_nonanonymous(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {"auth": {"allow-anonymous": False}}
        s = Session()
        s.username = "test"
        auth_plugin = AnonymousAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert ret


class TestFileAuthPlugin(unittest.TestCase):
    def setUp(self) -> None:
        self.loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()

    def test_allow(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {
            "auth": {
                "password-file": Path(__file__).parent / "passwd",
            },
        }
        s = Session()
        s.username = "user"
        s.password = "test"
        auth_plugin = FileAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert ret

    def test_wrong_password(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {
            "auth": {
                "password-file": Path(__file__).parent / "passwd",
            },
        }
        s = Session()
        s.username = "user"
        s.password = "wrong password"
        auth_plugin = FileAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert not ret

    def test_unknown_password(self) -> None:
        context = BaseContext()
        context.logger = logging.getLogger(__name__)
        context.config = {
            "auth": {
                "password-file": Path(__file__).parent / "passwd",
            },
        }
        s = Session()
        s.username = "some user"
        s.password = "some password"
        auth_plugin = FileAuthPlugin(context)
        ret = self.loop.run_until_complete(auth_plugin.authenticate(session=s))
        assert not ret



import asyncio
import logging
import tempfile
from pathlib import Path

import pytest
from passlib.context import CryptContext
from typer.testing import CliRunner

from amqtt.contexts import Action
from amqtt.contrib.auth_db.managers import TopicManager, UserManager
from amqtt.contrib.auth_db.models import PasswordHasher, AllowedTopic
from amqtt.contrib.auth_db.topic_mgr_cli import topic_app
from amqtt.contrib.auth_db.user_mgr_cli import user_app

runner = CliRunner()

@pytest.fixture
def password_hasher():
    pwd_hasher = PasswordHasher()
    pwd_hasher.crypt_context = CryptContext(schemes=["argon2", ], deprecated="auto")
    yield pwd_hasher


@pytest.fixture
def db_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        with tempfile.NamedTemporaryFile(mode='wb', delete=True) as tmp:
            yield Path(temp_dir) / f"{tmp.name}.db"


@pytest.fixture
def db_connection(db_file):
    test_db_connect = f"sqlite+aiosqlite:///{db_file}"
    yield test_db_connect


@pytest.fixture
@pytest.mark.asyncio
async def user_manager(password_hasher, db_connection):
    um = UserManager(db_connection)
    await um.db_sync()
    yield um


@pytest.fixture
@pytest.mark.asyncio
async def topic_manager(password_hasher, db_connection):
    tm = TopicManager(db_connection)
    await tm.db_sync()
    yield tm


@pytest.mark.parametrize("app,error_msg", [
    (user_app, "user cli"),
    (topic_app, "topic cli"),
])
def test_cli_mgr_no_params(app, error_msg):

    result = runner.invoke(app, [])
    assert result.exit_code == 0, f"{result.output}"


@pytest.mark.parametrize("app,error_msg", [
    (user_app, "user cli"),
    (topic_app, "topic cli"),
])
def test_cli_mgr_no_db_type(app, error_msg):
    result = runner.invoke(topic_app, ["sync"])
    assert result.exit_code == 2


@pytest.mark.parametrize("app,error_msg", [
    (user_app, "user cli"),
    (topic_app, "topic cli"),
])
def test_cli_mgr_no_db_username(app, error_msg, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(app, ["-d", "mysql", "sync"])
        assert result.exit_code == 1
        assert "DB access requires a username be provided." in caplog.text


@pytest.mark.parametrize("app,error_msg", [
    (user_app, "user cli"),
    (topic_app, "topic cli"),
])
def test_cli_mgr_db_not_installed(app, error_msg, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(app,
                               ["-d", "mysql", "-u", "mydbname", "sync",],
                               input="mydbpassword\n"
                               )
        assert result.exit_code == 1
        assert isinstance(result.exception, ModuleNotFoundError)


@pytest.mark.parametrize("app,error_msg", [
    (user_app, "user cli"),
    (topic_app, "topic cli"),
])
def test_cli_mgr_sync(db_file, app, error_msg, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "sync"
        ])
        assert result.exit_code == 0
        assert "Success: database synced." in caplog.text


@pytest.mark.parametrize("app,success_msg", [
    (user_app, "authentications"),
    (topic_app, "authorizations"),
])
def test_topic_empty_list(db_file, topic_manager, caplog, app, success_msg):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "list"
        ])
        assert result.exit_code == 0
        assert f"No client {success_msg} exist." in caplog.text


def test_user_mgr_list_clients(db_file, user_manager, caplog):
    async def init_user_auths():
        await user_manager.create_user_auth("client123", "randompassword")
        await user_manager.create_user_auth("client456", "randompassword")

    asyncio.run(init_user_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "list"
        ])
        assert result.exit_code == 0

        info_records = [record for record in caplog.records if record.levelname == "INFO"]

        assert 'client123' in info_records[0].message
        assert 'client456' in info_records[1].message


def test_topic_mgr_list_clients(db_file, topic_manager, caplog):
    async def init_topic_auths():
        await topic_manager.create_topic_auth('device123')
        await topic_manager.create_topic_auth('device456')
        await topic_manager.add_allowed_topic('device123', 'my/topic', Action.SUBSCRIBE)
        await topic_manager.add_allowed_topic('device456', 'my/topic', Action.PUBLISH)

    asyncio.run(init_topic_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(topic_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "list"
        ])
        assert result.exit_code == 0

        info_records = [record for record in caplog.records if record.levelname == "INFO"]

        assert 'device123' in info_records[0].message
        assert 'my/topic' in info_records[0].message
        assert 'device456' in info_records[1].message
        assert 'my/topic' in info_records[1].message


def test_user_mgr_add_auth_missing_param(db_file, user_manager, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "add"
        ])
        assert result.exit_code == 2


def test_add_allowed_topic_missing_param(db_file, topic_manager, caplog):

    with caplog.at_level(logging.INFO):
        result = runner.invoke(topic_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "add", "-c", "client123", "my/topic"
        ])
        assert result.exit_code == 2


def test_user_mgr_add_auth_missing_password(db_file, user_manager, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "add", "-c", 'client123'
        ], input=" \n")
        assert result.exit_code == 1
        assert "Error: client password cannot be empty." in caplog.text, caplog.text

    async def verify_add():
        user_auth = await user_manager.get_user_auth('client123')
        assert user_auth is None

    asyncio.run(verify_add())


def test_user_mgr_add_auth(db_file, user_manager, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "add", "-c", 'client123'
        ], input="dbpassword\nuserpassword\n")
        assert result.exit_code == 0
        assert "Success: created 'client123'" in caplog.text, caplog.text

    async def verify_add():
        user_auth = await user_manager.get_user_auth('client123')
        assert user_auth is not None

    asyncio.run(verify_add())


def test_add_allowed_topic(db_file, topic_manager, caplog):
    with caplog.at_level(logging.INFO):
        result = runner.invoke(topic_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "add",
            "-c", "client123",
            "-a", "publish",
            "my/topic"
        ])
        assert result.exit_code == 0
        assert "Success: topic 'my/topic' added to publish for 'client123'" in caplog.text

    async def verify_add():
        topic_auth = await topic_manager.get_topic_auth('client123')
        assert topic_auth is not None
        assert AllowedTopic('my/topic') in topic_auth.publish_acl

    asyncio.run(verify_add())


def test_remove_user_auth_mismatch(db_file, user_manager, caplog):
    async def init_user_auths():
        await user_manager.create_user_auth("client123", "randompassword")

    asyncio.run(init_user_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "rm",
            "-c", "device123",
        ])
        assert result.exit_code == 1, caplog.text
        assert "Error: client 'device123' does not exist." in caplog.text, result.output


def test_remove_allowed_topic_mismatch(db_file, topic_manager, caplog):
    async def init_topic_auths():
        await topic_manager.create_topic_auth('device123')
        await topic_manager.add_allowed_topic('device123', 'my/topic', Action.SUBSCRIBE)

    asyncio.run(init_topic_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(topic_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "rm",
            "-c", "device123",
            "-a", "publish",
            "my/topic"
        ])
        assert result.exit_code == 1, caplog.text
        assert "Error: topic 'my/topic' not in the publish allow list for device123." in caplog.text


def test_remove_user_auth_abort(db_file, user_manager, caplog):
    async def init_user_auths():
        await user_manager.create_user_auth("client123", "randompassword")

    asyncio.run(init_user_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "rm",
            "-c", "client123",
        ], input="N\n")
        assert result.exit_code == 0, caplog.text

    async def verify_user_exists():
        assert await user_manager.get_user_auth('client123') is not None

    asyncio.run(verify_user_exists())


def test_remove_user_auth_confirm(db_file, user_manager, caplog):
    async def init_user_auths():
        await user_manager.create_user_auth("client123", "randompassword")

    asyncio.run(init_user_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "rm",
            "-c", "client123",
        ], input="y\n")
        assert result.exit_code == 0, caplog.text
        assert "Success: 'client123' was removed." in caplog.text, result.output

    async def verify_user_removed():
        assert await user_manager.get_user_auth('client123') is None

    asyncio.run(verify_user_removed())


def test_remove_allowed_topic(db_file, topic_manager, caplog):
    async def init_topic_auth():
        await topic_manager.create_topic_auth('device123')
        await topic_manager.add_allowed_topic('device123', 'my/topic', Action.SUBSCRIBE)

    asyncio.run(init_topic_auth())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(topic_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "rm",
            "-c", "device123",
            "-a", "subscribe",
            "my/topic"
        ])
        assert result.exit_code == 0
        assert "Success: removed topic 'my/topic' from subscribe for 'device123'" in caplog.text, caplog.text


def test_user_mgr_change_password_empty(db_file, user_manager, caplog):
    async def init_user_auths():
        await user_manager.create_user_auth("client123", "randompassword")

    asyncio.run(init_user_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "pwd",
            "-c", "client123",
        ], input=" \n")
        assert result.exit_code == 1, caplog.text
        assert "Error: client password cannot be empty." in caplog.text, caplog.text


def test_user_mgr_change_password(db_file, user_manager, caplog):
    async def init_user_auths():
        await user_manager.create_user_auth("client123", "randompassword")
        user_auth = await user_manager.get_user_auth('client123')
        return user_auth._password_hash

    orig_pwd_hash = asyncio.run(init_user_auths())

    with caplog.at_level(logging.INFO):
        result = runner.invoke(user_app, [
            "-d", "sqlite",
            "-f", f"{db_file}",
            "pwd",
            "-c", "client123",
        ], input="myotherpassword\n")
        assert result.exit_code == 0, caplog.text
        assert "Success: client 'client123' password updated." in caplog.text, caplog.text

    async def verify_user_exists(pwd_hash):
        user_auth = await user_manager.get_user_auth('client123')
        assert user_auth is not None
        assert user_auth._password_hash != pwd_hash

    asyncio.run(verify_user_exists(orig_pwd_hash))

@pytest.mark.asyncio
async def test_user_mgr_cli():
    cmd = [
        "user_mgr",
        "--help"]

    proc = await asyncio.create_subprocess_shell(
        " ".join(cmd), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    await asyncio.sleep(0.2)
    stdout, stderr = await proc.communicate()

    assert proc.returncode == 0, f"user_mgr error code: {proc.returncode} - {stdout} - {stderr}"


@pytest.mark.asyncio
async def test_topic_mgr_cli():
    cmd = [
        "topic_mgr",
        "--help"]

    proc = await asyncio.create_subprocess_shell(
        " ".join(cmd), stdin=asyncio.subprocess.PIPE, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    await asyncio.sleep(0.2)
    stdout, stderr = await proc.communicate()

    assert proc.returncode == 0, f"topic_mgr error code: {proc.returncode} - {stdout} - {stderr}"




import asyncio
import logging

from amqtt.client import ConnectError, MQTTClient
from amqtt.mqtt.constants import QOS_1, QOS_2

"""
This sample shows how to publish messages to broker using different QOS
"""

logger = logging.getLogger(__name__)

config = {
    "will": {
        "topic": "/will/client",
        "message": b"Dead or alive",
        "qos": QOS_1,
        "retain": True,
    },
}


async def test_coro1() -> None:
    client = MQTTClient()
    await client.connect("mqtt://localhost:1883/")
    tasks = [
        asyncio.ensure_future(client.publish("a/b", b"TEST MESSAGE WITH QOS_0")),
        asyncio.ensure_future(client.publish("a/b", b"TEST MESSAGE WITH QOS_1", qos=QOS_1)),
        asyncio.ensure_future(client.publish("a/b", b"TEST MESSAGE WITH QOS_2", qos=QOS_2)),
    ]
    await asyncio.wait(tasks)
    logger.info("test_coro1 messages published")
    await client.disconnect()


async def test_coro2() -> None:
    try:
        client = MQTTClient(config={"auto_reconnect": False, "connection_timeout": 1})
        await client.connect("mqtt://localhost:1884/")
        await client.publish("a/b", b"TEST MESSAGE WITH QOS_0", qos=0x00)
        await client.publish("a/b", b"TEST MESSAGE WITH QOS_1", qos=0x01)
        await client.publish("a/b", b"TEST MESSAGE WITH QOS_2", qos=0x02)
        logger.info("test_coro2 messages published")
        await client.disconnect()
    except ConnectError:
        logger.info("Connection failed", exc_info=True)


def __main__():

    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)

    asyncio.run(test_coro1())
    asyncio.run(test_coro2())

if __name__ == "__main__":
    __main__()



import asyncio
import math

import pytest
import time
from hamcrest import equal_to, assert_that, close_to, has_key, instance_of, has_entry



from amqtt.contrib.shadows import ShadowPlugin, ShadowOperation
from amqtt.contrib.shadows.states import State, StateDocument, calculate_delta_update, calculate_iota_update, \
    MetaTimestamp


@pytest.mark.parametrize("topic,client_id,shadow_name,message_type,is_match", [
    ('$shadow/myclientid/myshadow/get', 'myclientid', 'myshadow', ShadowOperation.GET, True),
    ('$shadow/myshadow/get', '', '', '', False)
])
def test_shadow_topic_match(topic, client_id, shadow_name, message_type, is_match):

    # broker_context = BrokerContext(broker=Broker())
    # shadow_plugin = ShadowPlugin(context=broker_context)
    shadow_topic = ShadowPlugin.shadow_topic_match(topic)
    if is_match:
        assert shadow_topic.device_id == client_id
        assert shadow_topic.name == shadow_name
        assert shadow_topic.message_op in ShadowOperation
        assert shadow_topic.message_op == message_type
    else:
        assert shadow_topic is None


@pytest.mark.asyncio
async def test_state_add():

    cur_time = math.floor(time.time())

    data = {
         'state':{
             'desired': {
                 'item1': 'value1a',
                 'item2': 'value2a'
             },
             'reported': {
                 'item1': 'value1a',
                 'item2': 'value2b'
             }
         }
    }

    meta = {
        'metadata': {
            'desired': {
                'item1': 10,
                'item2': 20
            },
            'reported': {
                'item1': 11,
                'item2': 21
            }
        }
    }

    data_state = State.from_dict(data['state'])
    meta_state = State.from_dict(meta['metadata'])

    state_document_one = StateDocument(state=data_state, metadata=meta_state)
    await asyncio.sleep(2)

    data_update = {
         'state':{
             'desired': {
                 'item2': 'value2a'
             },
             'reported': {
                 'item1': 'value1c',
                 'item2': 'value2c'
             }
         }
    }

    state_document_two = StateDocument.from_dict(data_update)

    final_doc = state_document_one + state_document_two

    assert final_doc.state.desired['item1'] == 'value1a'
    assert final_doc.metadata.desired['item1'] == 10

    assert final_doc.state.desired['item2'] == 'value2a'
    assert final_doc.metadata.desired['item2'] > cur_time

    assert final_doc.state.reported['item1'] == 'value1c'
    assert final_doc.metadata.reported['item1'] > cur_time
    assert final_doc.state.reported['item1'] == 'value1c'
    assert final_doc.metadata.reported['item1'] > cur_time


def test_state_from_dict() -> None:

    state_dict = {
        'desired': {'keyA': 'valueA', 'keyB': 'valueB'},
        'reported': {'keyC': 'valueC', 'keyD': 'valueD'}
    }

    state = State.from_dict(state_dict)

    assert_that(state.desired['keyA'], equal_to('valueA'))
    assert_that(state.desired['keyB'], equal_to('valueB'))
    assert_that(state.reported['keyC'], equal_to('valueC'))
    assert_that(state.reported['keyD'], equal_to('valueD'))


def test_state_doc_from_dict() -> None:
    now = int(time.time())

    state_dict = {
        'state': {
            'desired': {'keyA': 'valueA', 'keyB': 'valueB'},
            'reported': {'keyC': 'valueC', 'keyD': 'valueD'}
        }
    }

    state_doc = StateDocument.from_dict(state_dict)

    assert_that(state_doc.state.desired['keyA'], equal_to('valueA'))
    assert_that(state_doc.state.desired['keyB'], equal_to('valueB'))

    assert_that(state_doc.metadata.desired, has_key('keyA'))  # noqa
    assert_that(state_doc.metadata.desired, has_key('keyB'))  # noqa

    assert_that(state_doc.metadata.desired['keyA'], instance_of(MetaTimestamp))  # noqa
    assert_that(state_doc.metadata.desired['keyB'], instance_of(MetaTimestamp))  # noqa
    assert_that(state_doc.metadata.desired['keyA'], close_to(now, 0))  # noqa
    assert_that(state_doc.metadata.desired['keyA'], equal_to(state_doc.metadata.desired['keyB']))


def test_state_doc_including_meta() -> None:
    now = int(time.time())
    state1 = State(
            desired={'keyA': 'valueA', 'keyB': 'valueB'},
            reported={'keyC': 'valueC', 'keyD': 'valueD'}
        )
    meta1 = State(
        desired={'keyA': now - 100, 'keyB': now - 110},
        reported={'keyC': now - 90, 'keyD': now - 120}
    )

    state_doc1 = StateDocument(
        state=state1,
        metadata=meta1
    )

    state2 = State(
        desired={'keyA': 'valueA', 'keyB': 'valueB'},
        reported={'keyC': 'valueC', 'keyD': 'valueD'}
    )
    meta2 = State(
        desired={'keyA': now - 5, 'keyB': now - 5},
        reported={'keyC': now - 5, 'keyD': now - 5}
    )

    state_doc2 = StateDocument(
        state=state2,
        metadata=meta2
    )

    new_doc = state_doc1 + state_doc2

    assert_that(new_doc.metadata.desired['keyA'], equal_to(now - 5))
    assert_that(new_doc.metadata.reported['keyC'], equal_to(now - 5))


def test_state_doc_plus_new_key_update() -> None:
    now = int(time.time())
    state = State(
            desired={'keyA': 'valueA', 'keyB': 'valueB'},
            reported={'keyC': 'valueC', 'keyD': 'valueD'}
        )
    meta = State(
        desired={'keyA': now - 100, 'keyB': now - 110},
        reported={'keyC': now - 90, 'keyD': now - 120}
    )

    state_doc = StateDocument(
        state=state,
        metadata=meta
    )

    update_dict = {'state': {'reported': {'keyE': 'valueE', 'keyF': 'valueF'}}}
    update = StateDocument.from_dict(update_dict)

    next_doc = state_doc + update

    assert_that(next_doc.state.reported, has_key('keyC'))
    assert_that(next_doc.metadata.reported['keyC'], equal_to(now - 90))
    assert_that(next_doc.metadata.reported['keyD'], equal_to(now - 120))

    assert_that(next_doc.state.reported, has_key('keyE'))
    assert_that(next_doc.state.reported, has_key('keyF'))
    assert_that(next_doc.metadata.reported['keyE'], close_to(now, 1))
    assert_that(next_doc.metadata.reported['keyF'], close_to(now, 1))


def test_state_with_updated_keys() -> None:

    now = int(time.time())
    state = State(
            desired={'keyA': 'valueA', 'keyB': 'valueB'},
            reported={'keyC': 'valueC', 'keyD': 'valueD'}
        )
    meta = State(
        desired={'keyA': now - 100, 'keyB': now - 110},
        reported={'keyC': now - 90, 'keyD': now - 120}
    )

    state_doc = StateDocument(
        state=state,
        metadata=meta
    )

    update_dict = {'state': {'reported': {'keyD': 'valueD'}}}
    update = StateDocument.from_dict(update_dict)

    next_doc = state_doc + update

    assert_that(next_doc.state.reported, has_key('keyC'))
    assert_that(next_doc.state.reported, has_key('keyD'))
    assert_that(next_doc.metadata.reported['keyC'], equal_to(now - 90))
    assert_that(next_doc.metadata.reported['keyD'], close_to(now, 1))


def test_update_with_empty_initial_state() -> None:
    now = int(time.time())

    prev_doc = StateDocument.from_dict({})

    state = State(
            desired={'keyA': 'valueA', 'keyB': 'valueB'},
            reported={'keyC': 'valueC', 'keyD': 'valueD'}
        )

    state_doc = StateDocument(state=state)
    new_doc = prev_doc + state_doc

    assert_that(state_doc.state.desired, equal_to(new_doc.state.desired))
    assert_that(state_doc.state.reported, equal_to(new_doc.state.reported))
    assert_that(state_doc.metadata.reported, has_key('keyC'))
    assert_that(state_doc.metadata.reported['keyC'], close_to(now, 1))


def test_update_with_clearing_key() -> None:
    state = State(
            desired={'keyA': 'valueA', 'keyB': 'valueB'},
            reported={'keyC': 'valueC', 'keyD': 'valueD'}
        )

    state_doc = StateDocument(state=state)

    update_doc = StateDocument.from_dict({'state': {'reported': {'keyC': None}}})

    new_doc = state_doc + update_doc

    assert_that(new_doc.state.reported, has_entry(equal_to('keyC'), equal_to(None)))


def test_empty_desired_state() -> None:

    state_doc = StateDocument.from_dict({
        'state': {
            'reported': {
                'items': ['value1', 'value2', 'value3']
            }
        }
    })

    diff = calculate_delta_update(state_doc.state.desired, state_doc.state.reported)
    assert diff == {}


def test_empty_reported_state() -> None:

    state_doc = StateDocument.from_dict({
        'state': {
            'desired': {
                'items': ['value1', 'value2', 'value3']
            }
        }
    })

    diff = calculate_delta_update(state_doc.state.desired, state_doc.state.reported)
    assert diff == {'items': ['value1', 'value2', 'value3']}


def test_matching_desired_reported_state() -> None:

    state_doc = StateDocument.from_dict({
        'state': {
            'desired': {
                'items': ['value1', 'value2', 'value3']
            },
            'reported': {
                'items': ['value1', 'value2', 'value3']
            }
        }
    })

    diff = calculate_delta_update(state_doc.state.desired, state_doc.state.reported)
    assert diff == {}


def test_out_of_order_list() -> None:

    state_doc = StateDocument.from_dict({
        'state': {
            'desired': {
                'items': ['value1', 'value2', 'value3']
            },
            'reported': {
                'items': ['value2', 'value1', 'value3']
            }
        }
    })

    diff = calculate_delta_update(state_doc.state.desired, state_doc.state.reported)
    assert diff == { 'items': ['value1', 'value2', 'value3'] }

def test_states_with_connector_transaction() -> None:
    state_doc = StateDocument.from_dict({
        'state': {
            'desired': {},
            'reported': {"transaction": False, "transactionId": "5678", "tag": "ghijk"}
        }
    })

    diff = calculate_iota_update(state_doc.state.desired, state_doc.state.reported)
    assert diff == {"transaction": None, "transactionId": None, "tag": None}


def test_extra_reported_into_desired_with_overlap() -> None:
    state_doc = StateDocument.from_dict({
        'state': {
            'desired': {"connectors": [1, 2, 3]},
            'reported': {"status": None, "heartbeat": "2025-02-07T04:16:51.431Z", "connectors": [1, 2, 3],
                "module_version": "2.0.1", "restartTime": "2025-02-07T03:16:51.431Z"}
        }})

    diff = calculate_iota_update(state_doc.state.desired, state_doc.state.reported)
    assert diff == {"status": None, "heartbeat": None, "module_version": None, "restartTime": None}




import logging

from dataclasses import dataclass

from amqtt.plugins.base import BasePlugin, BaseAuthPlugin, BaseTopicPlugin
from amqtt.contexts import BaseContext, Action

from amqtt.session import Session

logger = logging.getLogger(__name__)


class TestSimplePlugin(BasePlugin):

    def __init__(self, context: BaseContext):
        super().__init__(context)


class TestConfigPlugin(BasePlugin):

    def __init__(self, context: BaseContext):
        super().__init__(context)

    @dataclass
    class Config:
        option1: int
        option2: str
        option3: int = 20


class TestCoroErrorPlugin(BaseAuthPlugin):

    def authenticate(self, *, session: Session) -> bool | None:
        return True


class TestAuthPlugin(BaseAuthPlugin):

    async def authenticate(self, *, session: Session) -> bool | None:
        return True


class TestNoAuthPlugin(BaseAuthPlugin):


    async def authenticate(self, *, session: Session) -> bool | None:
        return False


class TestAllowTopicPlugin(BaseTopicPlugin):

    def __init__(self, context: BaseContext):
        super().__init__(context)

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool:
        return True


class TestBlockTopicPlugin(BaseTopicPlugin):

    def __init__(self, context: BaseContext):
        super().__init__(context)

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool:
        logger.debug("topic filtering plugin is returning false")
        return False



import logging
import sys

from amqtt.errors import MQTTError

logger = logging.getLogger(__name__)


def main() -> None:
    """Run the auth db cli."""
    try:
        from amqtt.contrib.auth_db.user_mgr_cli import user_app  # pylint: disable=import-outside-toplevel
    except ImportError:
        logger.critical("optional 'contrib' library is missing, please install: `pip install amqtt[contrib]`")
        sys.exit(1)

    from amqtt.contrib.auth_db.user_mgr_cli import user_app  # pylint: disable=import-outside-toplevel
    try:
        user_app()
    except ModuleNotFoundError as mnfe:
        logger.critical(f"Please install database-specific dependencies: {mnfe}")
        sys.exit(1)
    except ValueError as ve:
        if "greenlet" in f"{ve}":
            logger.critical("Please install database-specific dependencies: 'greenlet'")
            sys.exit(1)
        logger.critical(f"Unknown error: {ve}")
        sys.exit(1)

    except MQTTError as me:
        logger.critical(f"could not execute command: {me}")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""INIT."""
import re
from typing import Any, Optional


class TopicMatcher:

    _instance: Optional["TopicMatcher"] = None

    def __init__(self) -> None:
        if not hasattr(self, "_topic_filter_matchers"):
            self._topic_filter_matchers: dict[str, re.Pattern[str]] = {}

    def __new__(cls, *args: list[Any], **kwargs: dict[str, Any]) -> "TopicMatcher":
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def is_topic_allowed(self, topic: str, a_filter: str) -> bool:
        if topic.startswith("$") and (a_filter.startswith(("+", "#"))):
            return False

        if "#" not in a_filter and "+" not in a_filter:
            # if filter doesn't contain wildcard, return exact match
            return a_filter == topic

        # else use regex (re.compile is an expensive operation, store the matcher for future use)
        if a_filter not in self._topic_filter_matchers:
            self._topic_filter_matchers[a_filter] = re.compile(re.escape(a_filter)
                                                               .replace("\\#", "?.*")
                                                               .replace("\\+", "[^/]*")
                                                               .lstrip("?"))
        match_pattern = self._topic_filter_matchers[a_filter]
        return bool(match_pattern.fullmatch(topic))

    def are_topics_allowed(self, topic: str, many_filters: list[str]) -> bool:

        return any(self.is_topic_allowed(topic, a_filter) for a_filter in many_filters)



from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from sqlalchemy import String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from amqtt.contexts import Action
from amqtt.contrib import DataClassListJSON
from amqtt.plugins import TopicMatcher

if TYPE_CHECKING:
    from passlib.context import CryptContext


logger = logging.getLogger(__name__)

matcher = TopicMatcher()


@dataclass
class AllowedTopic:
    topic: str

    def __contains__(self, item: Union[str, "AllowedTopic"]) -> bool:
        """Determine `in`."""
        return self.__eq__(item)

    def __eq__(self, item: object) -> bool:
        """Determine `==` or `!=`."""
        if isinstance(item, str):
            return matcher.is_topic_allowed(item, self.topic)
        if isinstance(item, AllowedTopic):
            return item.topic == self.topic
        msg = "AllowedTopic can only be compared to another AllowedTopic or string."
        raise AttributeError(msg)

    def __str__(self) -> str:
        """Display topic."""
        return self.topic

    def __repr__(self) -> str:
        """Display topic."""
        return self.topic


class PasswordHasher:
    """singleton to initialize the CryptContext and then use it elsewhere in the code."""

    _instance: Optional["PasswordHasher"] = None

    def __init__(self) -> None:

        if not hasattr(self, "_crypt_context"):
            self._crypt_context: CryptContext | None = None

    def __new__(cls, *args: list[Any], **kwargs: dict[str, Any]) -> "PasswordHasher":
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    @property
    def crypt_context(self) -> "CryptContext":
        if not self._crypt_context:
            msg = "CryptContext is empty"
            raise ValueError(msg)
        return self._crypt_context

    @crypt_context.setter
    def crypt_context(self, value: "CryptContext") -> None:
        self._crypt_context = value


class Base(DeclarativeBase):
    pass


class UserAuth(Base):
    __tablename__ = "user_auth"

    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(String, unique=True)
    _password_hash: Mapped[str] = mapped_column("password_hash", String(128))

    publish_acl: Mapped[list[AllowedTopic]] = mapped_column(DataClassListJSON(AllowedTopic), default=list)
    subscribe_acl: Mapped[list[AllowedTopic]] = mapped_column(DataClassListJSON(AllowedTopic), default=list)
    receive_acl: Mapped[list[AllowedTopic]] = mapped_column(DataClassListJSON(AllowedTopic), default=list)

    @hybrid_property
    def password(self) -> None:
        msg = "Password is write-only"
        raise AttributeError(msg)

    @password.inplace.setter  # type: ignore[arg-type]
    def _password_setter(self, plain_password: str) -> None:
        self._password_hash = PasswordHasher().crypt_context.hash(plain_password)

    def verify_password(self, plain_password: str) -> bool:
        return bool(PasswordHasher().crypt_context.verify(plain_password, self._password_hash))

    def __str__(self) -> str:
        """Display client id and password hash."""
        return f"'{self.username}' with password hash: {self._password_hash}"


class TopicAuth(Base):
    __tablename__ = "topic_auth"

    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(String, unique=True)

    publish_acl: Mapped[list[AllowedTopic]] = mapped_column(DataClassListJSON(AllowedTopic), default=list)
    subscribe_acl: Mapped[list[AllowedTopic]] = mapped_column(DataClassListJSON(AllowedTopic), default=list)
    receive_acl: Mapped[list[AllowedTopic]] = mapped_column(DataClassListJSON(AllowedTopic), default=list)

    def get_topic_list(self, action: Action) -> list[AllowedTopic]:
        return cast("list[AllowedTopic]", getattr(self, f"{action}_acl"))

    def __str__(self) -> str:
        """Display client id and password hash."""
        return f"""'{self.username}':
\tpublish: {self.publish_acl}, subscribe: {self.subscribe_acl}, receive: {self.receive_acl}
"""



import asyncio
import logging

from amqtt.client import ConnectError, MQTTClient
from amqtt.mqtt.constants import QOS_1

"""
This sample shows how to publish messages to broker running either `samples/broker_acl.py`
 or `samples/broker_taboo.py`. 
"""

logger = logging.getLogger(__name__)


async def test_coro() -> None:
    try:
        client = MQTTClient()
        await client.connect("mqtt://0.0.0.0:1883")
        await client.publish("data/classified", b"TOP SECRET", qos=QOS_1)
        await client.publish("data/memes", b"REAL FUN", qos=QOS_1)
        await client.publish("repositories/amqtt/master", b"NEW STABLE RELEASE", qos=QOS_1)
        await client.publish(
            "repositories/amqtt/devel",
            b"THIS NEEDS TO BE CHECKED",
            qos=QOS_1,
        )
        await client.publish("calendar/amqtt/releases", b"NEW RELEASE", qos=QOS_1)
        logger.info("messages published")
        await client.disconnect()
    except ConnectError:
        logger.exception("ERROR: Connection failed")


def __main__():

    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)

    asyncio.run(test_coro())

if __name__ == "__main__":
    __main__()



import asyncio
import logging
import secrets
from typing import Any
import unittest

from amqtt.adapters import StreamReaderAdapter, StreamWriterAdapter
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from amqtt.mqtt.protocol.handler import ProtocolHandler
from amqtt.mqtt.puback import PubackPacket
from amqtt.mqtt.pubcomp import PubcompPacket
from amqtt.mqtt.publish import PublishPacket
from amqtt.mqtt.pubrec import PubrecPacket
from amqtt.mqtt.pubrel import PubrelPacket
from amqtt.plugins.manager import PluginManager
from amqtt.session import IncomingApplicationMessage, OutgoingApplicationMessage, Session

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=formatter)
log = logging.getLogger(__name__)


def rand_packet_id():
    return secrets.randbelow(65536)


def adapt(reader, writer):
    return StreamReaderAdapter(reader), StreamWriterAdapter(writer)


class ProtocolHandlerTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.plugin_manager = PluginManager("amqtt.test.plugins", context=None)

    def tearDown(self):
        self.loop.close()

    def test_init_handler(self):

        async def test_coro() -> None:
            try:
                Session()
                handler = ProtocolHandler(self.plugin_manager)
                assert handler.session is None
                assert handler._loop is self.loop
                self.check_empty_waiters(handler)
                future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        future: asyncio.Future[Any] = asyncio.Future()
        self.loop.run_until_complete(test_coro())
        exception = future.exception()
        if exception:
            raise exception

    def test_start_stop(self):
        async def server_mock(reader, writer) -> None:
            writer.close()  # python 3.12 requires an explicit close
            await writer.wait_closed()

        async def test_coro() -> None:
            try:
                s = Session()
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                handler = ProtocolHandler(self.plugin_manager)
                handler.attach(s, reader_adapted, writer_adapted)
                await self.start_handler(handler, s)
                await self.stop_handler(handler, s)
                future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        future: asyncio.Future[Any] = asyncio.Future()
        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_publish_qos0(self):
        async def server_mock(reader, writer) -> None:
            try:
                packet = await PublishPacket.from_stream(reader)
                assert packet.variable_header.topic_name == "/topic"
                assert packet.qos == QOS_0
                assert packet.packet_id is None
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()

            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                s = Session()
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                handler = ProtocolHandler(self.plugin_manager)
                handler.attach(s, reader_adapted, writer_adapted)
                await self.start_handler(handler, s)
                message = await handler.mqtt_publish(
                    "/topic",
                    b"test_data",
                    QOS_0,
                    False,
                )
                assert isinstance(message, OutgoingApplicationMessage)
                assert message.publish_packet is not None
                assert message.puback_packet is None
                assert message.pubrec_packet is None
                assert message.pubrel_packet is None
                assert message.pubcomp_packet is None
                await self.stop_handler(handler, s)
                future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        future: asyncio.Future[Any] = asyncio.Future()
        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_publish_qos1(self):
        self.handler: ProtocolHandler | None = None

        async def server_mock(reader, writer) -> None:
            packet = await PublishPacket.from_stream(reader)
            try:
                assert packet.variable_header.topic_name == "/topic"
                assert packet.qos == QOS_1
                assert packet.packet_id is not None
                assert packet.packet_id in self.session.inflight_out
                assert self.handler is not None
                assert packet.packet_id in self.handler._puback_waiters
                puback = PubackPacket.build(packet.packet_id)
                await puback.to_stream(writer)
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()
            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.start_handler(self.handler, self.session)
                message = await self.handler.mqtt_publish(
                    "/topic",
                    b"test_data",
                    QOS_1,
                    False,
                )
                assert isinstance(message, OutgoingApplicationMessage)
                assert message.publish_packet is not None
                assert message.puback_packet is not None
                assert message.pubrec_packet is None
                assert message.pubrel_packet is None
                assert message.pubcomp_packet is None
                await self.stop_handler(self.handler, self.session)
                if not future.done():
                    future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        future: asyncio.Future[Any] = asyncio.Future()

        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_publish_qos2(self):
        async def server_mock(reader, writer) -> None:
            try:
                packet = await PublishPacket.from_stream(reader)
                assert packet.topic_name == "/topic"
                assert packet.qos == QOS_2
                assert packet.packet_id is not None
                assert packet.packet_id in self.session.inflight_out
                assert self.handler is not None
                assert packet.packet_id in self.handler._pubrec_waiters
                pubrec = PubrecPacket.build(packet.packet_id)
                await pubrec.to_stream(writer)

                await PubrelPacket.from_stream(reader)
                assert packet.packet_id in self.handler._pubcomp_waiters
                pubcomp = PubcompPacket.build(packet.packet_id)
                await pubcomp.to_stream(writer)
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()
            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.start_handler(self.handler, self.session)
                message = await self.handler.mqtt_publish(
                    "/topic",
                    b"test_data",
                    QOS_2,
                    False,
                )
                assert isinstance(message, OutgoingApplicationMessage)
                assert message.publish_packet is not None
                assert message.puback_packet is None
                assert message.pubrec_packet is not None
                assert message.pubrel_packet is not None
                assert message.pubcomp_packet is not None
                await self.stop_handler(self.handler, self.session)
                if not future.done():
                    future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        future: asyncio.Future[Any] = asyncio.Future()

        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_receive_qos0(self):
        async def server_mock(reader, writer) -> None:
            packet = PublishPacket.build(
                "/topic",
                b"test_data",
                rand_packet_id(),
                False,
                QOS_0,
                False,
            )
            await packet.to_stream(writer)
            writer.close()  # python 3.12 requires an explicit close
            await writer.wait_closed()

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.start_handler(self.handler, self.session)
                message = await self.handler.mqtt_deliver_next_message()
                assert isinstance(message, IncomingApplicationMessage)
                assert message.publish_packet is not None
                assert message.puback_packet is None
                assert message.pubrec_packet is None
                assert message.pubrel_packet is None
                assert message.pubcomp_packet is None
                await self.stop_handler(self.handler, self.session)
                future.set_result(True)

            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        future: asyncio.Future[Any] = asyncio.Future()
        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_receive_qos1(self):
        async def server_mock(reader, writer) -> None:
            try:
                packet = PublishPacket.build(
                    "/topic",
                    b"test_data",
                    rand_packet_id(),
                    False,
                    QOS_1,
                    False,
                )
                await packet.to_stream(writer)
                puback = await PubackPacket.from_stream(reader)
                assert puback is not None
                assert packet.packet_id == puback.packet_id
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()
            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.start_handler(self.handler, self.session)
                message = await self.handler.mqtt_deliver_next_message()
                assert isinstance(message, IncomingApplicationMessage)
                assert message.publish_packet is not None
                assert message.puback_packet is not None
                assert message.pubrec_packet is None
                assert message.pubrel_packet is None
                assert message.pubcomp_packet is None
                await self.stop_handler(self.handler, self.session)
                future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        future: asyncio.Future[Any] = asyncio.Future()
        self.event = asyncio.Event()
        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_receive_qos2(self):
        async def server_mock(reader, writer) -> None:
            try:
                packet = PublishPacket.build(
                    "/topic",
                    b"test_data",
                    rand_packet_id(),
                    False,
                    QOS_2,
                    False,
                )
                await packet.to_stream(writer)
                pubrec = await PubrecPacket.from_stream(reader)
                assert pubrec is not None
                assert packet.packet_id == pubrec.packet_id
                assert self.handler is not None
                assert packet.packet_id in self.handler._pubrel_waiters
                pubrel = PubrelPacket.build(packet.packet_id)
                await pubrel.to_stream(writer)
                pubcomp = await PubcompPacket.from_stream(reader)
                assert pubcomp is not None
                assert packet.packet_id == pubcomp.packet_id
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()
            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.start_handler(self.handler, self.session)
                message = await self.handler.mqtt_deliver_next_message()
                assert isinstance(message, IncomingApplicationMessage)
                assert message.publish_packet is not None
                assert message.puback_packet is None
                assert message.pubrec_packet is not None
                assert message.pubrel_packet is not None
                assert message.pubcomp_packet is not None
                await self.stop_handler(self.handler, self.session)
                future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        future: asyncio.Future[Any] = asyncio.Future()
        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    async def start_handler(self, handler, session):
        self.check_empty_waiters(handler)
        self.check_no_message(session)
        await handler.start()
        assert handler._reader_ready

    async def stop_handler(self, handler, session):
        await handler.stop()
        assert handler._reader_stopped
        self.check_empty_waiters(handler)
        self.check_no_message(session)

    def check_empty_waiters(self, handler):
        assert not handler._puback_waiters
        assert not handler._pubrec_waiters
        assert not handler._pubrel_waiters
        assert not handler._pubcomp_waiters

    def check_no_message(self, session):
        assert not session.inflight_out
        assert not session.inflight_in

    def test_publish_qos1_retry(self):
        async def server_mock(reader, writer) -> None:
            packet = await PublishPacket.from_stream(reader)
            try:
                assert packet.topic_name == "/topic"
                assert packet.qos == QOS_1
                assert packet.packet_id is not None
                assert packet.packet_id in self.session.inflight_out
                assert self.handler is not None
                assert packet.packet_id in self.handler._puback_waiters
                puback = PubackPacket.build(packet.packet_id)
                await puback.to_stream(writer)
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()
            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.handler.start()
                await self.stop_handler(self.handler, self.session)
                if not future.done():
                    future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        message = OutgoingApplicationMessage(1, "/topic", QOS_1, b"test_data", False)
        message.publish_packet = PublishPacket.build(
            "/topic",
            b"test_data",
            rand_packet_id(),
            False,
            QOS_1,
            False,
        )
        self.session.inflight_out[1] = message
        future: asyncio.Future[Any] = asyncio.Future()

        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception

    def test_publish_qos2_retry(self):
        async def server_mock(reader, writer) -> None:
            try:
                packet = await PublishPacket.from_stream(reader)
                assert packet.topic_name == "/topic"
                assert packet.qos == QOS_2
                assert packet.packet_id is not None
                assert packet.packet_id in self.session.inflight_out
                assert self.handler is not None
                assert packet.packet_id in self.handler._pubrec_waiters
                pubrec = PubrecPacket.build(packet.packet_id)
                await pubrec.to_stream(writer)

                await PubrelPacket.from_stream(reader)
                assert packet.packet_id in self.handler._pubcomp_waiters
                pubcomp = PubcompPacket.build(packet.packet_id)
                await pubcomp.to_stream(writer)
                writer.close()  # python 3.12 requires an explicit close
                await writer.wait_closed()

            except Exception as ae:
                future.set_exception(ae)

        async def test_coro() -> None:
            try:
                reader, writer = await asyncio.open_connection("127.0.0.1", 8888)
                reader_adapted, writer_adapted = adapt(reader, writer)
                self.handler = ProtocolHandler(self.plugin_manager)
                self.handler.attach(self.session, reader_adapted, writer_adapted)
                await self.handler.start()
                await self.stop_handler(self.handler, self.session)
                if not future.done():
                    future.set_result(True)
            except Exception as ae:
                future.set_exception(ae)

        self.handler = None
        self.session = Session()
        message = OutgoingApplicationMessage(1, "/topic", QOS_2, b"test_data", False)
        message.publish_packet = PublishPacket.build(
            "/topic",
            b"test_data",
            rand_packet_id(),
            False,
            QOS_2,
            False,
        )
        self.session.inflight_out[1] = message
        future: asyncio.Future[Any] = asyncio.Future()

        coro = asyncio.start_server(server_mock, "127.0.0.1", 8888)
        server = self.loop.run_until_complete(coro)
        self.loop.run_until_complete(test_coro())
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        exception = future.exception()
        if exception:
            raise exception



import asyncio
import unittest

from amqtt.adapters import BufferReader
from amqtt.mqtt.packet import PacketIdVariableHeader
from amqtt.mqtt.suback import SubackPacket, SubackPayload


class SubackPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b"\x90\x06\x00\x0a\x00\x01\x02\x80"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(SubackPacket.from_stream(stream))
        assert message.payload.return_codes[0] == SubackPayload.RETURN_CODE_00
        assert message.payload.return_codes[1] == SubackPayload.RETURN_CODE_01
        assert message.payload.return_codes[2] == SubackPayload.RETURN_CODE_02
        assert message.payload.return_codes[3] == SubackPayload.RETURN_CODE_80

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = SubackPayload(
            [
                SubackPayload.RETURN_CODE_00,
                SubackPayload.RETURN_CODE_01,
                SubackPayload.RETURN_CODE_02,
                SubackPayload.RETURN_CODE_80,
            ],
        )
        suback = SubackPacket(variable_header=variable_header, payload=payload)
        out = suback.to_bytes()
        assert out == b"\x90\x06\x00\n\x00\x01\x02\x80"



from typing import Any


class AMQTTError(Exception):
    """aMQTT base exception."""


class MQTTError(Exception):
    """Base class for all errors referring to MQTT specifications."""


class CodecError(Exception):
    """Exceptions thrown by packet encode/decode functions."""


class NoDataError(Exception):
    """Exceptions thrown by packet encode/decode functions."""


class ZeroLengthReadError(NoDataError):
    def __init__(self) -> None:
        super().__init__("Decoding a string of length zero.")


class BrokerError(Exception):
    """Exceptions thrown by broker."""


class PluginError(Exception):
    """Exceptions thrown when loading or initializing a plugin."""


class PluginImportError(PluginError):
    """Exceptions thrown when loading plugin."""


class PluginCoroError(PluginError):
    """Exceptions thrown when loading a plugin with a non-async call method."""


class PluginInitError(PluginError):
    """Exceptions thrown when initializing plugin."""

    def __init__(self, plugin: Any) -> None:
        super().__init__(f"Plugin init failed: {plugin!r}")


class ClientError(Exception):
    """Exceptions thrown by client."""


class ConnectError(ClientError):
    """Exceptions thrown by client connect."""

    return_code: int | None = None


class ProtocolHandlerError(Exception):
    """Exceptions thrown by protocol handle."""


class PluginLoadError(Exception):
    """Exception thrown when loading a plugin."""



import asyncio
import unittest
import pytest

from amqtt.adapters import BufferReader
from amqtt.errors import AMQTTError
from amqtt.mqtt import MQTTFixedHeader, PUBLISH
from amqtt.mqtt.pubcomp import PacketIdVariableHeader, PubcompPacket


class PubcompPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b"\x70\x02\x00\x0a"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(PubcompPacket.from_stream(stream))
        assert message.variable_header.packet_id == 10

    def test_to_bytes(self):
        variable_header = PacketIdVariableHeader(10)
        publish = PubcompPacket(variable_header=variable_header)
        out = publish.to_bytes()
        assert out == b"p\x02\x00\n"

def test_incorrect_fixed_header():
    header = MQTTFixedHeader(PUBLISH, 0x00)
    with pytest.raises(AMQTTError):
        _ = PubcompPacket(fixed=header)


@pytest.mark.parametrize("prop", [
    "packet_id"
])
def test_empty_variable_header(prop):
    packet = PubcompPacket()

    with pytest.raises(ValueError):
        assert getattr(packet, prop) is not None

    with pytest.raises(ValueError):
        assert setattr(packet, prop, "a value")



import asyncio
import logging
from pathlib import Path
import sqlite3
import pytest
import aiosqlite

from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy import select

from amqtt.broker import Broker, BrokerContext, RetainedApplicationMessage
from amqtt.client import MQTTClient
from amqtt.mqtt.constants import QOS_1
from amqtt.contrib.persistence import SessionDBPlugin, Subscription, StoredSession, RetainedMessage
from amqtt.session import Session

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=formatter)

logger = logging.getLogger(__name__)

@pytest.fixture
async def db_file():
    db_file = Path(__file__).parent / "amqtt.db"
    if db_file.exists():
        raise NotImplementedError("existing db file found, should it be cleaned up?")

    yield db_file
    if db_file.exists():
        db_file.unlink()

@pytest.fixture
async def broker_context():
    cfg = {
        'listeners': { 'default': {'type': 'tcp', 'bind': 'localhost:1883' }},
        'plugins': {}
    }

    context = BrokerContext(broker=Broker(config=cfg))
    yield context

@pytest.fixture
async def db_session_factory(db_file):
    engine = create_async_engine(f"sqlite+aiosqlite:///{str(db_file)}")
    factory = async_sessionmaker(engine, expire_on_commit=False)
    yield factory



@pytest.mark.asyncio
async def test_initialize_tables(db_file, broker_context):

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    assert db_file.exists()

    conn = sqlite3.connect(str(db_file))
    cursor = conn.cursor()
    table_name = 'stored_sessions'
    cursor.execute(f"PRAGMA table_info({table_name});")
    rows = cursor.fetchall()
    column_names = [row[1] for row in rows]
    assert len(column_names) > 1

@pytest.mark.asyncio
async def test_create_stored_session(db_file, broker_context, db_session_factory):

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    async with db_session_factory() as db_session:
        async with db_session.begin():
            stored_session = await session_db_plugin._get_or_create_session(db_session, 'test_client_1')
            assert stored_session.client_id == 'test_client_1'

    async with aiosqlite.connect(str(db_file)) as db:
        async with await db.execute("SELECT * FROM stored_sessions") as cursor:
            async for row in cursor:
                assert row[1] == 'test_client_1'

@pytest.mark.asyncio
async def test_get_stored_session(db_file, broker_context, db_session_factory):
    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    async with aiosqlite.connect(str(db_file)) as db:
        sql = """INSERT INTO stored_sessions (
    client_id, clean_session, will_flag,
    will_qos, keep_alive,
    retained, subscriptions
) VALUES (
    'test_client_1',
    1,
    0,
    1,
    60,
    '[]',
    '[{"topic":"sensors/#","qos":1}]'
)"""
        await db.execute(sql)
        await db.commit()

    async with db_session_factory() as db_session:
        async with db_session.begin():
            stored_session = await session_db_plugin._get_or_create_session(db_session, 'test_client_1')
            assert stored_session.subscriptions == [Subscription(topic='sensors/#', qos=1)]


@pytest.mark.asyncio
async def test_update_stored_session(db_file, broker_context, db_session_factory):
    broker_context.config = SessionDBPlugin.Config(file=db_file)

    # create session for client id (without subscription)
    await broker_context.add_subscription('test_client_1', None, None)

    session = broker_context.get_session('test_client_1')
    assert session is not None
    session.clean_session = False

    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    # initialize with stored client session
    async with aiosqlite.connect(str(db_file)) as db:
        sql = """INSERT INTO stored_sessions (
    client_id, clean_session, will_flag,
    will_qos, keep_alive,
    retained, subscriptions
) VALUES (
    'test_client_1',
    1,
    0,
    1,
    60,
    '[]',
    '[{"topic":"sensors/#","qos":1}]'
)"""
        await db.execute(sql)
        await db.commit()

    await session_db_plugin.on_broker_client_subscribed(client_id='test_client_1', topic='my/topic', qos=2)

    # verify that the stored session has been updated with the new subscription
    has_stored_session = False
    async with aiosqlite.connect(str(db_file)) as db:
        async with await db.execute("SELECT * FROM stored_sessions") as cursor:
            async for row in cursor:
                assert row[1] == 'test_client_1'
                assert row[-1] == '[{"topic": "sensors/#", "qos": 1}, {"topic": "my/topic", "qos": 2}]'
                has_stored_session = True
    assert has_stored_session, "stored session wasn't updated"


@pytest.mark.asyncio
async def test_client_connected_with_clean_session(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    session = Session()
    session.client_id = 'test_client_connected'
    session.is_anonymous = False
    session.clean_session = True

    await session_db_plugin.on_broker_client_connected(client_id='test_client_connected', client_session=session)

    async with aiosqlite.connect(str(db_file)) as db_conn:
        db_conn.row_factory = sqlite3.Row

        async with await db_conn.execute("SELECT * FROM stored_sessions") as cursor:
            assert len(await cursor.fetchall()) == 0


@pytest.mark.asyncio
async def test_client_connected_anonymous_session(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    session = Session()
    session.is_anonymous = True
    session.client_id = 'test_client_connected'

    await session_db_plugin.on_broker_client_connected(client_id='test_client_connected', client_session=session)

    async with aiosqlite.connect(str(db_file)) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory
        async with await db_conn.execute("SELECT * FROM stored_sessions") as cursor:
            assert len(await cursor.fetchall()) == 0


@pytest.mark.asyncio
async def test_client_connected_and_stored_session(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    session = Session()
    session.client_id = 'test_client_connected'
    session.is_anonymous = False
    session.clean_session = False
    session.will_flag = True
    session.will_qos = 1
    session.will_topic = 'my/will/topic'
    session.will_retain = False
    session.will_message = b'test connected client has a last will (and testament) message'
    session.keep_alive = 42

    await session_db_plugin.on_broker_client_connected(client_id='test_client_connected', client_session=session)

    has_stored_session = False
    async with aiosqlite.connect(str(db_file)) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        async with await db_conn.execute("SELECT * FROM stored_sessions") as cursor:
            for row in await cursor.fetchall():
                assert row['client_id'] == 'test_client_connected'
                assert row['clean_session'] == False
                assert row['will_flag'] == True
                assert row['will_qos'] == 1
                assert row['will_topic'] == 'my/will/topic'
                assert row['will_retain'] == False
                assert row['will_message'] == b'test connected client has a last will (and testament) message'
                assert row['keep_alive'] == 42
                has_stored_session = True
    assert has_stored_session, "client session wasn't stored"


@pytest.mark.asyncio
async def test_repopulate_stored_sessions(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    async with aiosqlite.connect(str(db_file)) as db:
        sql = """INSERT INTO stored_sessions (
        client_id, clean_session, will_flag,
        will_qos, keep_alive,
        retained, subscriptions
    ) VALUES (
        'test_client_1',
        1,
        0,
        1,
        60,
        '[{"topic":"sensors/#","data":"this message is retained when client reconnects","qos":1}]',
        '[{"topic":"sensors/#","qos":1}]'
    )"""
        await db.execute(sql)
        await db.commit()

    await session_db_plugin.on_broker_post_start()

    session = broker_context.get_session('test_client_1')
    assert session is not None
    assert session.retained_messages.qsize() == 1

    assert 'sensors/#' in broker_context._broker_instance._subscriptions

    # ugly: b/c _subscriptions is a list of dictionaries of tuples
    assert broker_context._broker_instance._subscriptions['sensors/#'][0][1] == 1


@pytest.mark.asyncio
async def test_client_retained_message(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    # add a session to the broker
    await broker_context.add_subscription('test_retained_client', None, None)

    # update the session so that it's retained
    session = broker_context.get_session('test_retained_client')
    assert session is not None
    session.is_anonymous = False
    session.clean_session = False
    session.transitions.disconnect()

    retained_message = RetainedApplicationMessage(source_session=session, topic='my/retained/topic', data=b'retain message for disconnected client', qos=2)
    await session_db_plugin.on_broker_retained_message(client_id='test_retained_client', retained_message=retained_message)

    async with db_session_factory() as db_session:
        async with db_session.begin():
            stmt = select(StoredSession).filter(StoredSession.client_id == 'test_retained_client')
            stored_session = await db_session.scalar(stmt)
            assert stored_session is not None
            assert len(stored_session.retained) > 0
            assert RetainedMessage(topic='my/retained/topic', data='retained message', qos=2)


@pytest.mark.asyncio
async def test_topic_retained_message(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file, clear_on_shutdown=False)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    # add a session to the broker
    await broker_context.add_subscription('test_retained_client', None, None)

    # update the session so that it's retained
    session = broker_context.get_session('test_retained_client')
    assert session is not None
    session.is_anonymous = False
    session.clean_session = False
    session.transitions.disconnect()

    retained_message = RetainedApplicationMessage(source_session=session, topic='my/retained/topic', data=b'retained message', qos=2)
    await session_db_plugin.on_broker_retained_message(client_id=None, retained_message=retained_message)

    has_stored_message = False
    async with aiosqlite.connect(str(db_file)) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        async with await db_conn.execute("SELECT * FROM stored_messages") as cursor:
            # assert(len(await cursor.fetchall()) > 0)
            for row in await cursor.fetchall():
                assert row['topic'] == 'my/retained/topic'
                assert row['data'] == b'retained message'
                assert row['qos'] == 2
                has_stored_message = True

    assert has_stored_message, "retained topic message wasn't stored"


@pytest.mark.asyncio
async def test_topic_clear_retained_message(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    # add a session to the broker
    await broker_context.add_subscription('test_retained_client', None, None)

    # update the session so that it's retained
    session = broker_context.get_session('test_retained_client')
    assert session is not None
    session.is_anonymous = False
    session.clean_session = False
    session.transitions.disconnect()

    retained_message = RetainedApplicationMessage(source_session=session, topic='my/retained/topic', data=b'', qos=0)
    await session_db_plugin.on_broker_retained_message(client_id=None, retained_message=retained_message)

    async with aiosqlite.connect(str(db_file)) as db_conn:
        db_conn.row_factory = sqlite3.Row

        async with await db_conn.execute("SELECT * FROM stored_messages") as cursor:
            assert(len(await cursor.fetchall()) == 0)


@pytest.mark.asyncio
async def test_restoring_retained_message(db_file, broker_context, db_session_factory) -> None:

    broker_context.config = SessionDBPlugin.Config(file=db_file)
    session_db_plugin = SessionDBPlugin(broker_context)
    await session_db_plugin.on_broker_pre_start()

    stmts = ("INSERT INTO stored_messages VALUES(1,'my/retained/topic1',X'72657461696e6564206d657373616765',2)",
             "INSERT INTO stored_messages VALUES(2,'my/retained/topic2',X'72657461696e6564206d65737361676532',2)",
             "INSERT INTO stored_messages VALUES(3,'my/retained/topic3',X'72657461696e6564206d65737361676533',2)")

    async with aiosqlite.connect(str(db_file)) as db:
        for stmt in stmts:
            await db.execute(stmt)
        await db.commit()

    await session_db_plugin.on_broker_post_start()

    assert len(broker_context.retained_messages) == 3
    assert 'my/retained/topic1' in broker_context.retained_messages
    assert 'my/retained/topic2' in broker_context.retained_messages
    assert 'my/retained/topic3' in broker_context.retained_messages


@pytest.fixture
def db_config(db_file):
    return {
        'listeners': {
            'default': {
                'type': 'tcp',
                'bind': '127.0.0.1:1883'
            }
        },
        'plugins': {
            'amqtt.plugins.authentication.AnonymousAuthPlugin': {'allow_anonymous': False},
            'amqtt.contrib.persistence.SessionDBPlugin': {
                'file': db_file,
                'clear_on_shutdown': False

            }
        }
    }




@pytest.mark.asyncio
async def test_broker_client_no_cleanup(db_file, db_config) -> None:

    b1 = Broker(config=db_config)
    await b1.start()
    await asyncio.sleep(0.1)

    c1 = MQTTClient(client_id='test_client1', config={'auto_reconnect':False})
    await c1.connect("mqtt://myUsername@127.0.0.1:1883", cleansession=False)

    # test that this message is retained for the topic upon restore
    await c1.publish("my/retained/topic", b'retained message for topic my/retained/topic', retain=True)
    await asyncio.sleep(0.2)
    await c1.disconnect()
    await b1.shutdown()

    # new broker should load the previous broker's db file since clean_on_shutdown is false in config
    b2 = Broker(config=db_config)
    await b2.start()
    await asyncio.sleep(0.1)

    # upon subscribing to topic with retained message, it should be received
    c2 = MQTTClient(client_id='test_client2', config={'auto_reconnect':False})
    await c2.connect("mqtt://myOtherUsername@localhost:1883", cleansession=False)
    await c2.subscribe([
        ('my/retained/topic', QOS_1)
    ])
    msg = await c2.deliver_message(timeout_duration=1)
    assert msg is not None
    assert msg.topic == "my/retained/topic"
    assert msg.data == b'retained message for topic my/retained/topic'

    await c2.disconnect()
    await b2.shutdown()


@pytest.mark.asyncio
async def test_broker_client_retain_subscription(db_file, db_config) -> None:

    b1 = Broker(config=db_config)
    await b1.start()
    await asyncio.sleep(0.1)

    c1 = MQTTClient(client_id='test_client1', config={'auto_reconnect':False})
    await c1.connect("mqtt://myUsername@127.0.0.1:1883", cleansession=False)

    # test to make sure the subscription is re-established upon reconnection after broker restart (clear_on_shutdown = False)
    ret = await c1.subscribe([
        ('my/offline/topic', QOS_1)
    ])
    assert ret == [QOS_1,]
    await asyncio.sleep(0.2)
    await c1.disconnect()
    await asyncio.sleep(0.1)
    await b1.shutdown()

    # new broker should load the previous broker's db file
    b2 = Broker(config=db_config)
    await b2.start()
    await asyncio.sleep(0.1)

    # client1's subscription should have been restored, so when it connects, it should receive this message
    c2 = MQTTClient(client_id='test_client2', config={'auto_reconnect':False})
    await c2.connect("mqtt://myOtherUsername@localhost:1883", cleansession=False)
    await c2.publish('my/offline/topic', b'standard message to be retained for offline clients')
    await asyncio.sleep(0.1)
    await c2.disconnect()

    await c1.reconnect(cleansession=False)
    await asyncio.sleep(0.1)
    msg = await c1.deliver_message(timeout_duration=2)
    assert msg is not None
    assert msg.topic == "my/offline/topic"
    assert msg.data == b'standard message to be retained for offline clients'

    await c1.disconnect()
    await b2.shutdown()


@pytest.mark.asyncio
async def test_broker_client_retain_message(db_file, db_config) -> None:
    """test to make sure that the retained message because client1 is offline,
     gets sent when back online after broker restart."""

    b1 = Broker(config=db_config)
    await b1.start()
    await asyncio.sleep(0.1)

    c1 = MQTTClient(client_id='test_client1', config={'auto_reconnect':False})
    await c1.connect("mqtt://myUsername@127.0.0.1:1883", cleansession=False)

    # subscribe to a topic with QOS_1 so that we receive messages, even if we're disconnected when sent
    ret = await c1.subscribe([
        ('my/offline/topic', QOS_1)
    ])
    assert ret == [QOS_1,]
    await asyncio.sleep(0.2)
    # go offline
    await c1.disconnect()
    await asyncio.sleep(0.1)

    # another client sends a message to previously subscribed to topic
    c2 = MQTTClient(client_id='test_client2', config={'auto_reconnect':False})
    await c2.connect("mqtt://myOtherUsername@localhost:1883", cleansession=False)

    # this message should be delivered after broker stops and restarts (and client connects)
    await c2.publish('my/offline/topic', b'standard message to be retained for offline clients')
    await asyncio.sleep(0.1)
    await c2.disconnect()
    await asyncio.sleep(0.1)
    await b1.shutdown()

    # new broker should load the previous broker's db file since we declared clear_on_shutdown = False in config
    b2 = Broker(config=db_config)
    await b2.start()
    await asyncio.sleep(0.1)

    # when first client reconnects, it should receive the message that had been previously retained for it
    await c1.reconnect(cleansession=False)
    await asyncio.sleep(0.1)
    msg = await c1.deliver_message(timeout_duration=2)
    assert msg is not None
    assert msg.topic == "my/offline/topic"
    assert msg.data == b'standard message to be retained for offline clients'

    # client should also receive a message if send on this topic
    await c1.publish("my/offline/topic", b"online message should also be received")
    await asyncio.sleep(0.1)
    msg = await c1.deliver_message(timeout_duration=2)
    assert msg is not None
    assert msg.topic == "my/offline/topic"
    assert msg.data == b'online message should also be received'

    await c1.disconnect()
    await b2.shutdown()



import asyncio
import logging
from importlib.metadata import EntryPoint
from logging.config import dictConfig
from unittest.mock import patch

import pytest

from amqtt.broker import Broker
from amqtt.client import MQTTClient
from amqtt.mqtt.constants import QOS_0

dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'verbose',
        }
    },
    'loggers': {
        'transitions': {
            'level': 'WARNING',
        }
    }
})

# logging.basicConfig(level=logging.DEBUG, format=formatter)

logger = logging.getLogger(__name__)



all_sys_topics = [
    '$SYS/broker/version',
    '$SYS/broker/load/bytes/received',
    '$SYS/broker/load/bytes/sent',
    '$SYS/broker/messages/received',
    '$SYS/broker/messages/sent',
    '$SYS/broker/time',
    '$SYS/broker/uptime',
    '$SYS/broker/uptime/formatted',
    '$SYS/broker/clients/connected',
    '$SYS/broker/clients/disconnected',
    '$SYS/broker/clients/maximum',
    '$SYS/broker/clients/total',
    '$SYS/broker/messages/inflight',
    '$SYS/broker/messages/inflight/in',
    '$SYS/broker/messages/inflight/out',
    '$SYS/broker/messages/inflight/stored',
    '$SYS/broker/messages/publish/received',
    '$SYS/broker/messages/publish/sent',
    '$SYS/broker/messages/retained/count',
    '$SYS/broker/messages/subscriptions/count',
    '$SYS/broker/heap/size',
    '$SYS/broker/heap/maximum',
    '$SYS/broker/cpu/percent',
    '$SYS/broker/cpu/maximum',
]



# test broker sys
@pytest.mark.asyncio
async def test_broker_sys_plugin_deprecated_config() -> None:

    sys_topic_flags = {sys_topic:False for sys_topic in all_sys_topics}

    class MockEntryPoints:

        def select(self, group) -> list[EntryPoint]:
            match group:
                case 'tests.mock_plugins':
                    return [
                            EntryPoint(name='broker_sys', group='tests.mock_plugins', value='amqtt.plugins.sys.broker:BrokerSysPlugin'),
                            EntryPoint(name='auth_anonymous', group='test.mock_plugins', value='amqtt.plugins.authentication:AnonymousAuthPlugin'),
                        ]
                case _:
                    return list()


    with patch("amqtt.plugins.manager.entry_points", side_effect=MockEntryPoints) as mocked_mqtt_publish:

        config = {
            "listeners": {
                "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
            },
            'sys_interval': 1,
            'auth': {
                'allow_anonymous': True
            }
        }

        broker = Broker(plugin_namespace='tests.mock_plugins', config=config)
        await broker.start()
        client = MQTTClient()
        await client.connect("mqtt://127.0.0.1:1883/")
        await client.subscribe([("$SYS/#", QOS_0),])
        await client.publish('test/topic', b'my test message')
        await asyncio.sleep(2)
        sys_msg_count = 0
        try:
            while sys_msg_count < 30:
                message = await client.deliver_message(timeout_duration=1)
                if '$SYS' in message.topic:
                    sys_msg_count += 1
                    assert message.topic in sys_topic_flags
                    sys_topic_flags[message.topic] = True

        except asyncio.TimeoutError:
            logger.debug(f"TimeoutError after {sys_msg_count} messages")

        await client.disconnect()
        await broker.shutdown()

        assert sys_msg_count > 1

        assert all(sys_topic_flags.values()), f'topic not received: {[ topic for topic, flag in sys_topic_flags.items() if not flag ]}'


@pytest.mark.asyncio
async def test_broker_sys_plugin_config() -> None:

    sys_topic_flags = {sys_topic:False for sys_topic in all_sys_topics}

    config = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'plugins': [
            {'amqtt.plugins.authentication.AnonymousAuthPlugin': {'allow_anonymous': True}},
            {'amqtt.plugins.sys.broker.BrokerSysPlugin': {'sys_interval': 1}},
        ]
    }

    broker = Broker(plugin_namespace='tests.mock_plugins', config=config)
    await broker.start()
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1:1883/")
    await client.subscribe([("$SYS/#", QOS_0), ])
    await client.publish('test/topic', b'my test message')
    await asyncio.sleep(2)
    sys_msg_count = 0
    try:
        while sys_msg_count < 30:
            message = await client.deliver_message(timeout_duration=1)
            if '$SYS' in message.topic:
                sys_msg_count += 1
                assert message.topic in sys_topic_flags
                sys_topic_flags[message.topic] = True

    except asyncio.TimeoutError:
        logger.debug(f"TimeoutError after {sys_msg_count} messages")

    await client.disconnect()
    await broker.shutdown()

    assert sys_msg_count > 1

    assert all(
        sys_topic_flags.values()), f'topic not received: {[topic for topic, flag in sys_topic_flags.items() if not flag]}'
